﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion


using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Analyzers {
  [Item("CrossoverPerformanceAnalyzer", "An operator that analyzes the performance of crossovers.")]
  [StorableClass]
  public class CrossoverPerformanceAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string GenerationsParameterName = "Generations";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion


    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<ItemArray<Permutation>> ParentsParameter {
      get { return (ScopeTreeLookupParameter<Permutation>)Parameters["Parents"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ParentsQualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["ParentsQuality"]; }
    }
    public ILookupParameter<Permutation> ChildParameter {
      get { return ((LookupParameter<Permutation>)Parameters["Child"]); }
    }
    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    #endregion

    ScatterPlot plot;
    DataRow dtRow;
    int cnt = 0;

    [StorableConstructor]
    private CrossoverPerformanceAnalyzer(bool deserializing) : base(deserializing) { }
    private CrossoverPerformanceAnalyzer(CrossoverPerformanceAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public CrossoverPerformanceAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));

      Parameters.Add(new ScopeTreeLookupParameter<Permutation>("Parents", "The parent permutations which have been crossed."));
      ParentsParameter.ActualName = "TSPTour";

      Parameters.Add(new LookupParameter<Permutation>("Child", "The child permutation resulting from the crossover."));
      ChildParameter.ActualName = "TSPTour";

      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("ParentsQuality", "The quality of the parent solutions."));
      ParentsQualityParameter.ActualName = "TSPTourLength";

      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The evaluated quality of the child solution."));
      QualityParameter.ActualName = "TSPTourLength";
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new CrossoverPerformanceAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var qualityParent1 = ParentsQualityParameter.ActualValue.First().Value;
      var qualityParent2 = ParentsQualityParameter.ActualValue.Last().Value;
      Point2D<double> curPoint;

      double worseQuality = qualityParent1 > qualityParent2 ? qualityParent1 : qualityParent2;
      curPoint = new Point2D<double>(cnt++, worseQuality - QualityParameter.ActualValue.Value);

      string curGenStr = GenerationsParameter.ActualValue.Value.ToString();
      ScatterPlotDataRow row;

      if (!Results.ContainsKey("Scatter Plot")) {
        InitializePlot();
        Results.Add(new Result("Scatter Plot", plot));
        Results.Add(new Result("Scatter Plot History", new ScatterPlotHistory()));
        cnt = 0;

        DataTable dt = new DataTable("Average Crossover Performance");
        dtRow = new DataRow("Average Crossover Performance per Generation");
        dt.Rows.Add(dtRow);
        Results.Add(new Result("Average Crossover Performance", dt));
      }

      if (!plot.Rows.ContainsKey(curGenStr)) {
        if (GenerationsParameter.ActualValue.Value != 0) {
          double avg = plot.Rows[(GenerationsParameter.ActualValue.Value - 1).ToString()].Points.Average(x => x.Y);
          dtRow.Values.Add(avg);

          ((ScatterPlotHistory)Results["Scatter Plot History"].Value).Add(plot);
          InitializePlot();
          Results["Scatter Plot"].Value = plot;
          cnt = 0;
        }

        var points = new List<Point2D<double>>();
        points.Add(curPoint);
        row = new ScatterPlotDataRow(curGenStr, null, points);
        row.VisualProperties.PointStyle = ScatterPlotDataRowVisualProperties.ScatterPlotDataRowPointStyle.Circle;
        row.VisualProperties.PointSize = 5;
        plot.Rows.Add(row);
      } else {
        plot.Rows[curGenStr].Points.Add(curPoint);
      }

      return base.Apply();
    }

    private void InitializePlot() {
      plot = new ScatterPlot("Crossover Performance", null);
      plot.VisualProperties.XAxisTitle = "Solution Index";
      plot.VisualProperties.YAxisTitle = "Absolut Quality Difference";
    }
  }
}
